#lang racket
(require redex/reduction-semantics
         "same-ty.rkt")

(test-equal (term (union () ())) (term ()))
(test-equal (term (union (A B C) (X Y Z))) (term (A B C X Y Z)))
(test-equal (term (union (A X B Y C) (X A Y B Z))) (term (A X B Y C Z)))
(test-equal (term (minus (A B C) C)) (term (A B)))
(test-equal (term (minus (A B C) D)) (term (A B C)))
(test-equal (term (ftv Int)) (term ()))
(test-equal (term (ftv (Int -> Int))) (term ()))
(test-equal (term (ftv (Int -> X))) (term (X)))
(test-equal (term (ftv (X -> X))) (term (X)))
(test-equal (term (ftv (∀ X (Y -> X)))) (term (Y)))
(test-equal (term (ftv (∀ Y (∀ X (Y -> X))))) (term ()))

(test-equal (term (not-in X (X))) #f)
(test-equal (term (not-in X (Y))) #t)
(test-equal (term (not-in X (A B C X Y Z))) #f)
(test-equal (term (not-in X (A B C Y Z))) #t)

(test-equal (find-prefix '()) 'X)
(test-equal (find-prefix '(X XXX)) 'XXXX)
(test-equal (find-prefix '(XXX X)) 'XXXX)

(test-equal (term (inc_X X0)) (term X1))
(test-equal (term (inc_X ABBA110)) (term ABBA111))

(test-equal (term (normalize Int X0 ())) 
            (term Int))
(test-equal (term (normalize * X0 ()))
            (term *))
(test-equal (term (normalize (Int -> Int) X0 ()))
            (term (Int -> Int)))
(test-equal (term (normalize (∀ X X) X0 ()))
            (term (∀ X0 X0)))

(test-equal (term (normalize ((∀ X X) -> (∀ X X)) X0 ()))
            (term ((∀ X0 X0) -> (∀ X0 X0))))
(test-equal (term (normalize (∀ X (∀ X X)) X0 ()))
            (term (∀ X0 (∀ X1 X1))))

(test-equal (term (same-ty (∀ Y1 (Y1 -> Y1)) (∀ Z (Z -> Z))))
            (term #t))
(test-equal (term (same-ty (∀ X (∀ Y (X -> Y)))
                           (∀ Y (∀ X (Y -> X)))))
            (term #t))
(test-equal (term (same-ty (∀ X (∀ Y (X -> Y)))
                           (∀ X (∀ Y (Y -> X)))))
            (term #f))
(test-equal (term (same-ty X X))
            (term #t))
(test-equal (term (same-ty X Y))
            (term #f))

(test-equal (term (subst-ty X * (∀ X X)))
            (term (∀ X X)))
(test-equal (term (subst-ty X * (∀ Y X)))
            (term (∀ X1 *)))
(test-equal (term (subst-ty X * (∀ Y (X -> Y))))
            (term (∀ X1 (* -> X1))))
(test-equal (term (subst-ty X Z (∀ Z X)))
            (term (∀ X1 Z)))

(test-equal (term (same-ty/Γ mt (∀ X X) (∀ Y Y)))
            (term #t))
(test-equal (term (same-ty/Γ (mt X Int) (X -> X) (Int -> Int)))
            (term #t))
(test-equal (term (same-ty/Γ (mt X Int) (Int -> Int) (X -> X)))
            (term #t))
(test-equal (term (same-ty/Γ (mt X Int) (X -> X) (X -> Int)))
            (term #t))
(test-equal (term (same-ty/Γ (mt X Int) (X -> X) (Bool -> Int)))
            (term #f))
(test-equal (term (same-ty/Γ (mt X Int) (X -> Int) (X -> X)))
            (term #t))
(test-equal (term (same-ty/Γ ((mt X Int) Y Bool)
                             ((X -> X) -> (Y -> Y))
                             ((Int -> Int) -> (Bool -> Bool))))
            (term #t))
(test-equal (term (same-ty/Γ ((mt X Int) Y Bool)
                             ((X -> X) -> (Y -> Y))
                             ((X -> Int) -> (Bool -> Bool))))
            (term #t))
(test-equal (term (same-ty/Γ ((mt X Int) Y Bool)
                             ((X -> X) -> (Y -> Y))
                             ((Int -> Int) -> (Y -> Bool))))
            (term #t))
(test-equal (term (same-ty/Γ (mt f (Int -> Int)) Int Int))
            (term #t))


(test-equal (term (wft mt Int)) #t)
(test-equal (term (wft mt *)) #t)
(test-equal (term (wft mt (Int -> Int))) #t)
(test-equal (term (wft mt (∀ X X))) #t)
(test-equal (term (wft mt (∀ X X))) #t)
(test-equal (term (wft ((mt X Int) Z) Z)) #t)
(test-equal (term (wft ((mt Z) x Z) Z)) #t)
(test-equal (term (wft ((mt Z) x Int) Z)) #t)
(test-equal (term (wft (((mt X Int) Z) x Z) Z)) #t)

(test-equal (term (close-ty mt X)) (term X))
(test-equal (term (close-ty mt (X -> Y))) (term (X -> Y)))
(test-equal (term (close-ty (mt X Int) (X -> Y))) (term (Int -> Y)))
(test-equal (term (close-ty (mt Y Int) (X -> Y))) (term (X -> Int)))
(test-equal (term (close-ty (mt X Int) (∀ X X))) (term (∀ X X)))
(test-equal (term (close-ty (mt Z Int) (∀ Y (∀ X ((X -> Y) -> Z))))) (term (∀ Y (∀ X ((X -> Y) -> Int)))))
(test-equal (term (close-ty ((mt X Int) Y X) X)) (term Int))
(test-equal (term (close-ty ((mt x Int) Y X) Y)) (term X))
(test-equal (term (close-ty (((mt x Int) Y X) X Bool) Y)) (term X))
(test-equal (term (close-ty (((mt X Int) Y X) X Bool) Y)) (term Int))

(test-results)
