#lang racket
(require redex/reduction-semantics)

(provide same-ty same-ty/Γ same-tys/Γ
         normalize inc_X find-prefix not-in
         ftv union minus
         wft lookup
         close-ty
         subst-ty)

;; this is code duplication (the grammar for types)
;; but avoiding it is probably not worth the effort
(define-language ty
  (ι Int Bool)
  ((A B C) ι (A -> B) * (∀ X A) X)
  (Γ mt  (Γ x A) (Γ X) (Γ X A))
  ((X Y Z) 
   (side-condition (name X variable-not-otherwise-mentioned)
                   (regexp-match #rx"^[A-Z]" (symbol->string (term X)))))
  ((x y z) 
   (side-condition (name x variable-not-otherwise-mentioned)
                   (regexp-match #rx"^[a-z]" (symbol->string (term x)))))
  (bool #t #f))

(define-metafunction ty
  ftv : A -> (X ...)
  [(ftv X) (X)]
  [(ftv (∀ X B)) (minus (ftv B) X)]
  [(ftv ι) ()]
  [(ftv (A -> B)) (union (ftv A) (ftv B))]
  [(ftv *) ()])

(define-metafunction ty
  union : (X ...) (X ...) -> (X ...)
  [(union (X_1 ... X_2 X_3 ...) (X_2 X_4 ...))
   (union (X_1 ... X_2 X_3 ...) (X_4 ...))]
  [(union (X_1 ...) (X_2 X_3 ...))
   (union (X_1 ... X_2) (X_3 ...))]
  [(union (X ...) ()) (X ...)])

(define-metafunction ty
  minus : (X ...) X -> (X ...)
  [(minus (X_1 ... X_2 X_3 ...) X_2) (X_1 ... X_3 ...)]
  [(minus (X_1 ...) X_2) (X_1 ...)])
 
(define-relation ty
  same-ty/Γ ⊂ Γ × A × A
  [(same-ty/Γ Γ A B)
   (same-ty (close-ty Γ A) (close-ty Γ B))])

;; this version of same-ty/Γ is an earlier
;; version that we don't need anymore (probably)
#;
(define-relation ty
  same-ty/Γ ⊂ Γ × A × A
  [(same-ty/Γ mt A B)
   (same-ty A B)]
  [(same-ty/Γ (Γ X) A B)
   (same-ty/Γ Γ A B)]
  [(same-ty/Γ (Γ x C) A B)
   (same-ty/Γ Γ A B)]
  [(same-ty/Γ (Γ X C) A B)
   (same-ty/Γ Γ A B)]
  [(same-ty/Γ (Γ X C) A B)
   (same-ty/Γ Γ (subst-ty X C A) B)]
  [(same-ty/Γ (Γ X C) A B)
   (same-ty/Γ Γ A (subst-ty X C B))])


#|


   A =α B
 ------------
  . |- A = B


   Γ |- A = B
 --------------
  Γ,X |- A = B
  

   Γ |- A = B
 --------------
  Γ,x:=C |- A = B


     Γ |- A = B
 -------------------
   Γ,X:=C |- A = B


   Γ |= A{X:=C} = B
 -------------------
   Γ,X:=C |- A = B


   Γ |= A = B{X:=C}
 -------------------
   Γ,X:=C |- A = B


|#

(define-relation ty
  same-tys/Γ ⊂ Γ × (A ...) × (A ...)
  [(same-tys/Γ Γ (A_1 A_2 ...) (B_1 B_2 ...))
   (same-ty/Γ Γ A_1 B_1)
   (same-tys/Γ Γ (A_2 ...) (B_2 ...))]
  [(same-tys/Γ Γ () ())])

(define-metafunction ty
  subst-ty : X A A -> A
  [(subst-ty X A ι) ι]
  [(subst-ty X A *) *]
  [(subst-ty X A (B_1 -> B_2))
   ((subst-ty X A B_1) -> (subst-ty X A B_2))]
  [(subst-ty X A X) A]
  [(subst-ty X A Y) Y]
  [(subst-ty X A (∀ X B)) (∀ X B)]
  [(subst-ty X A (∀ Y B))
   (∀ Z (subst-ty X A (rename Y Z B)))
   (where Z ,(variable-not-in (term B) (term X)))])

(define-metafunction ty
  rename : X Y A -> A
  [(rename X Y X) Y]
  [(rename X Y Z) Z]
  [(rename X Y ι) ι]
  [(rename X Y *) *]
  [(rename X Y (A -> B))
   ((rename X Y A) -> (rename X Y B))]
  [(rename X Y (∀ X B))
   (∀ Y (rename X Y B))]
  [(rename X Y (∀ Z B))
   (∀ Z (rename X Y B))])
  
(define-metafunction ty
  close-ty : Γ A -> A
  [(close-ty mt X) X]
  [(close-ty (Γ X) X) X]
  [(close-ty (Γ Y) X) (close-ty Γ X)]
  [(close-ty (Γ x A) X) (close-ty Γ X)]
  [(close-ty (Γ X A) X) (close-ty Γ A)]
  [(close-ty (Γ Y A) X) (close-ty Γ X)]
  [(close-ty Γ *) *]
  [(close-ty Γ ι) ι]
  [(close-ty Γ (A -> B)) ((close-ty Γ A) -> (close-ty Γ B))]
  [(close-ty Γ (∀ X A)) (∀ X (close-ty (Γ X) A))])

(define-metafunction ty
  same-ty : A A -> bool
  [(same-ty A B) 
   ,(let ([X0 (string->symbol (format "~a0" (find-prefix (term (A B)))))])
      (equal? (term (normalize A ,X0 ()))
              (term (normalize B ,X0 ()))))])

(define-metafunction ty
  normalize : A X ((X X) ...) -> A
  [(normalize X_1 X_next ((X_2 Y_2) ... (X_1 Y_1) (X_3 Y_3) ...))
   Y_1
   (side-condition (term (not-in X_1 (X_2 ...))))]
  [(normalize X X_next ((X_1 Y_1) ...))
   X]
  [(normalize (∀ X A) X_next ((X_1 Y_1) ...))
   (∀ X_next (normalize A (inc_X X_next) ((X X_next) (X_1 Y_1) ...)))]
  [(normalize ι X_next ((X_1 X_2) ...)) ι]
  [(normalize (A -> B) X_next ((X_1 X_2) ...))
   ((normalize A X_next ((X_1 X_2) ...))
    ->
    (normalize B X_next ((X_1 X_2) ...)))]
  [(normalize * X_next ((X_1 X_2) ...)) *])

(define-metafunction ty
  inc_X : X -> X
  [(inc_X X) 
   ,(let* ([m (regexp-match #rx"([^0-9]*)([0-9]+)" (format "~a" (term X)))]
           [prefix (list-ref m 1)]
           [num (list-ref m 2)])
      (string->symbol
       (format "~a~a" prefix (+ (string->number num) 1))))])


(define (find-prefix sexp)
  (let ([prefix "X"])
    (let loop ([sexp sexp])
      (cond
        [(pair? sexp)
         (loop (car sexp))
         (loop (cdr sexp))]
        [(symbol? sexp)
         (when (regexp-match (regexp (format "^~a" prefix)) (symbol->string sexp))
           (set! prefix (string-append prefix "X"))
           (loop sexp))]))
    (string->symbol prefix)))

(define-metafunction ty
  not-in : X (X ...) -> any
  [(not-in X_1 (X_2 ... X_1 X_3 ...)) #f]
  [(not-in X_1 (X_2 ...)) #t])

(define-relation ty
  wft ⊂ Γ × A
  [(wft (Γ X) X)]
  [(wft (Γ Y) X) (wft Γ X)]
  
  [(wft (Γ x A) X) (wft Γ X)]
  
  [(wft (Γ X A) X)]
  [(wft (Γ Y A) X) (wft Γ X)]
  
  [(wft Γ (∀ X B)) (wft (Γ X) B)]
  [(wft Γ (A -> B)) (wft Γ A) (wft Γ B)]
  [(wft Γ ι)]
  [(wft Γ *)])

(define-metafunction ty
  lookup : variable Γ -> any ;; A or X or (not-here!)
  [(lookup X (Γ X A)) A]
  [(lookup X (Γ X)) X]
  [(lookup x (Γ x A)) A]
  
  [(lookup variable (Γ Y)) (lookup variable Γ)]
  [(lookup variable (Γ Y A)) (lookup variable Γ)]
  [(lookup variable (Γ y A)) (lookup variable Γ)]
  [(lookup variable mt) (not-here!)])
  
