#lang racket
(require redex/reduction-semantics
         "stlc.rkt"
         "same-ty.rkt"
         "preservation.rkt")
(provide stlc-red-test
         stlc-preservation)

(test-equal (term (stlc-tc mt 1))
            (term Int))
(test-equal (term (stlc-tc mt true))
            (term Bool))
(test-equal (term (stlc-tc mt false))
            (term Bool))
(test-equal (term (stlc-tc mt (+ 1 2)))
            (term Int))
(test-equal (term (stlc-tc mt (+ 1 2 3)))
            (term illtyped))
(test-equal (term (stlc-tc mt (+ 1 false)))
            (term illtyped))
(test-equal (term (stlc-tc mt (λ x : Int false)))
            (term (Int -> Bool)))
(test-equal (term (stlc-tc mt (λ x : Int x)))
            (term (Int -> Int)))
(test-equal (term (stlc-tc mt ((λ f : (Int -> Int) (f 1))
                               (λ x : Int 3))))
            (term Int))
(test-equal (term (stlc-tc mt ((λ f : (Int -> Int) (f 1))
                               (λ x : Int true))))
            (term illtyped))
(test-equal (term (stlc-tc mt (λ x : Int (λ y : Bool y))))
            (term (Int -> (Bool -> Bool))))
(test-equal (term (stlc-tc mt (λ x : Int (λ y : Bool x))))
            (term (Int -> (Bool -> Int))))
(test-equal (term (stlc-tc mt (λ x : Int (λ y : Bool z))))
            (term illtyped))
(test-equal (term (stlc-tc mt (((if true) (λ z : Int 1)) (λ z : Int 0))))
            (term Int))

(define (stlc-red-test red)
  (test-->> red
            (term (+ 1 2))
            (term 3))
  (test-->> red
            (term (+ (+ 1 2) 3))
            (term 6))
  (test-->> red
            (term (- 2 1))
            (term 1))
  (test-->> red
            (term (+ (- 1 2) 3))
            (term 2))
  (test-->> red
            (term ((λ x : Int x) 1))
            (term 1))
  (test-->> red
            (term ((λ y : Int y) 1))
            (term 1))
  (test-->> red
            (term ((λ y : Int (- y 3)) (+ 1 2)))
            (term 0))
  
  (test-results))

(stlc-red-test stlc-red)

(define (stlc-preservation)
  (type-preservation (term 1) (term 1))
  (type-preservation (term (+ (+ 1 2) (+ 3 4))) (term 10))
  (type-preservation (term (= (+ 1 2) (+ 3 4))) (term false))
  (type-preservation (term (= (+ 1 2) (+ 2 1))) (term true))
  (type-preservation (term ((λ x : Int x) 1)) (term 1))
  (type-preservation (term (((λ x : (Int -> Int) x) (λ x : Int 3)) 4)) (term 3))
  (type-preservation (term (((λ x : Int (λ y : Int x)) 1) 2)) (term 1))
  (type-preservation (term (((λ x : Int (λ y : Int (+ x y))) 1) 2)) (term 3))
  (type-preservation (term (((if true) (λ z : Int 1)) (λ z : Int 0)))
                     (term 1))
  (type-preservation (term (((if (= (+ 1 1) 3)) (λ z : Int 1)) (λ z : Int 0)))
                     (term 0)))

(parameterize ([current-tc (λ (x) (term (stlc-tc mt ,x)))]
               [current-red stlc-red]
               [current-same-ty (λ (x y) (term (same-ty ,x ,y)))])
  (stlc-preservation)
  (show-steps stlc-red))
