#lang racket
(require "subst.rkt"
         "contract-nu.rkt"
         "nu.rkt"
         redex/reduction-semantics)

;; check if in[x:=a] is the same as out
;; also, make sure that in and out are 't's in contract-nu
;; and make sure that 'a' is either a 't' or an 'A'.
(define-syntax (try stx)
  (syntax-case stx ()
    [(_ x a in out)
     #`(begin (verify-term-or-type (term in))
              (verify-term-or-type (term out))
              (verify-term-or-type (term a))
              #,(syntax/loc stx
                  (test-equal (term (subst x a in))
                              (term out))))]))

(define contract-nu:t? (redex-match contract-nu t))
(define contract-nu:A? (redex-match contract-nu A))
(define nu:t? (redex-match nu t))
(define nu:A? (redex-match nu A))
(define (t? a) (or (contract-nu:t? a) (nu:t? a)))
(define (A? a) (or (contract-nu:A? a) (nu:A? a)))

(define (verify-term-or-type a)
  (unless (or (t? a)
              (A? a))
    (error 'verify-term "expected a 't' or an 'A', but got:\n  ~s" a)))

(try x 1 x 1)
(try x 1 y y)
(try x 1 (+ x y) (+ 1 y))
(try x (λ y : X z) (λ z : X x) (λ z1 : X (λ y : X z)))
(try x (λ y : X z) (Λ X (λ y : X x)) (Λ X1 (λ y1 : X1 (λ y : X z))))
(try X (Y -> Z) (∀ Z (∀ W (X -> (Z -> W)))) (∀ Z1 (∀ W2 ((Y -> Z) -> (Z1 -> W2)))))
(try X Z 
     (ν Z := Int ((1 : Z ⇒ p *) : * ⇒ q X))
     (ν Z1 := Int ((1 : Z1 ⇒ p *) : * ⇒ q Z)))
(try X Z
     (ν Z := Int (λ x : Z (λ y : X y)))
     (ν Z1 := Int (λ x : Z1 (λ y1 : Z y1))))
(try X Z (Λ Z (λ z : Z (λ x : X x))) (Λ Z1 (λ z : Z1 (λ x1 : Z x1))))
(test-results)
